#include <mitsuba/render/scene.h>
#include <mitsuba/hw/basicshader.h>
#include <mitsuba/core/fresolver.h>
#include <algorithm>

MTS_NAMESPACE_BEGIN
using namespace std;

const int ERF_SEG_NUM = 8;
const double ERF_SEG_COORD[ERF_SEG_NUM + 1] = {-10000000000.000000, -3.000000, -2.000000, -1.000000, 0.000000, 1.000000, 2.000000, 3.000000, 10000000000.000000};
const double ERF_SEG_COEFF[ERF_SEG_NUM][3] = {{4.418099e-25, 6.627149e-15, -0.999978},
                                              {7.771843e-03, 4.351486e-02, -0.939380},
                                              {1.883745e-01, 7.177449e-01, -0.313330},
                                              {3.965979e-01, 1.239299e+00, 0.000000},
                                              {-3.965979e-01, 1.239299e+00, 0.000000},
                                              {-1.883745e-01, 7.177449e-01, 0.313330},
                                              {-7.771843e-03, 4.351486e-02, 0.939380},
                                              {-4.418099e-25, 6.627149e-15, 0.999978}};

struct MipData {
	float y1a, y1b, y2a, y2b;
	MipData() {}
	MipData(float y1a, float y1b, float y2a, float y2b) : y1a(y1a), y1b(y1b), y2a(y2a), y2b(y2b) {}
};

double min4(float a, float b, float c, float d) {
	return min(min(min(a, b), c), d);
}

double max4(float a, float b, float c, float d) {
	return max(max(max(a, b), c), d);
}

class BumpNdf : public BSDF {
private:
	float *coeff[2][16];
	float *min_n1, *max_n1, *min_n2, *max_n2;
	MipData ***mipmap;
	int NORMAL_MAP_RESOLUTION;
	int HIGHEST_LEVEL;
public:
	float getCoeffA(int x, int y, int i, int j) const {
		return coeff[0][j * 4 + i][x * NORMAL_MAP_RESOLUTION + y];
	}

	float getCoeffB(int x, int y, int i, int j) const {
		return coeff[1][j * 4 + i][x * NORMAL_MAP_RESOLUTION + y];
	}

	pair<float, float> getNormal(float u, float v) const {
		// Bicubic interpolation
		double x = u * NORMAL_MAP_RESOLUTION;
		double y = v * NORMAL_MAP_RESOLUTION;
		int x1 = ((unsigned int) floor(x)) % NORMAL_MAP_RESOLUTION;
		int y1 = ((unsigned int) floor(y)) % NORMAL_MAP_RESOLUTION;

		double n1 = 0.0f, n2 = 0.0f;

		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				n1 += getCoeffA(x1, y1, i, j) * pow(x - floor(x), (double)i) * pow(y - floor(y), (double)j);
				n2 += getCoeffB(x1, y1, i, j) * pow(x - floor(x), (double)i) * pow(y - floor(y), (double)j);
			}
		}

		return make_pair((float)n1, (float)n2);
	}

	void read_coeff() {
		printf("Reading coefficents...");

		for (int i = 0; i < 2; i++)
			for (int j = 0; j < 16; j++)
				coeff[i][j] = new float[NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION];

		FILE *fp = fopen(Thread::getThread()->getFileResolver()->resolve(m_coeffFilename).string().c_str(), "rb");
		if (!fp)
			printf("\nCannot open coeff file!\n");
		for (int i = 0; i < 2; i++)
			for (int j = 0; j < 16; j++)
				fread(&coeff[i][j][0], sizeof(float), NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION, fp);
		fclose(fp);

		printf("OK\n");
	}

	void read_bound() {
		printf("Reading bounds...");

		min_n1 = new float[NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION];
		max_n1 = new float[NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION];
		min_n2 = new float[NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION];
		max_n2 = new float[NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION];

		if (!min_n1) 
			printf("\nCannot allocate memory!\n");
		if (!max_n1) 
			printf("\nCannot allocate memory!\n");
		if (!min_n2) 
			printf("\nCannot allocate memory!\n");
		if (!max_n2) 
			printf("\nCannot allocate memory!\n");


		FILE *fp = fopen(Thread::getThread()->getFileResolver()->resolve(m_boundFilename).string().c_str(), "rb");
		if (!fp)
			printf("\nCannot open bound file!\n");

		fread(max_n1, sizeof(float), NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION, fp);
		fread(min_n1, sizeof(float), NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION, fp);
		fread(max_n2, sizeof(float), NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION, fp);
		fread(min_n2, sizeof(float), NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION, fp);
		fclose(fp);

		printf("OK\n");
	}

	void generateMipmap() {
		printf("Generating mipmap...");

		mipmap = new MipData**[HIGHEST_LEVEL + 1];
		if (!mipmap) 
			printf("\nCannot allocate memory!\n");
		for (int i = 0; i <= HIGHEST_LEVEL; i++) {
			mipmap[i] = new MipData*[NORMAL_MAP_RESOLUTION >> i];
			if (!mipmap[i]) 
				printf("\nCannot allocate memory!\n");
			for (int j = 0; j < NORMAL_MAP_RESOLUTION >> i; j++) {
				mipmap[i][j] = new MipData[NORMAL_MAP_RESOLUTION >> i];
				if (!mipmap[i][j]) 
					printf("\nCannot allocate memory!\n");
			}
		}
		for (int i = 0; i < NORMAL_MAP_RESOLUTION; i++)
			for (int j = 0; j < NORMAL_MAP_RESOLUTION; j++) {
				mipmap[0][i][j].y1a = min_n1[i * NORMAL_MAP_RESOLUTION + j];
				mipmap[0][i][j].y1b = max_n1[i * NORMAL_MAP_RESOLUTION + j];
				mipmap[0][i][j].y2a = min_n2[i * NORMAL_MAP_RESOLUTION + j];
				mipmap[0][i][j].y2b = max_n2[i * NORMAL_MAP_RESOLUTION + j];
			}


		for (int level = 0; level < HIGHEST_LEVEL; level++) {
			int SIZE = NORMAL_MAP_RESOLUTION >> (level + 1);
			for (int i = 0; i < SIZE; i++) {
				for (int j = 0; j < SIZE; j++) {
					mipmap[level + 1][i][j].y1a = min4(mipmap[level][i * 2][j * 2].y1a, mipmap[level][i * 2 + 1][j * 2].y1a, mipmap[level][i * 2 + 1][j * 2 + 1].y1a, mipmap[level][i * 2][j * 2 + 1].y1a);
					mipmap[level + 1][i][j].y1b = max4(mipmap[level][i * 2][j * 2].y1b, mipmap[level][i * 2 + 1][j * 2].y1b, mipmap[level][i * 2 + 1][j * 2 + 1].y1b, mipmap[level][i * 2][j * 2 + 1].y1b);
					mipmap[level + 1][i][j].y2a = min4(mipmap[level][i * 2][j * 2].y2a, mipmap[level][i * 2 + 1][j * 2].y2a, mipmap[level][i * 2 + 1][j * 2 + 1].y2a, mipmap[level][i * 2][j * 2 + 1].y2a);
					mipmap[level + 1][i][j].y2b = max4(mipmap[level][i * 2][j * 2].y2b, mipmap[level][i * 2 + 1][j * 2].y2b, mipmap[level][i * 2 + 1][j * 2 + 1].y2b, mipmap[level][i * 2][j * 2 + 1].y2b);
				}
			}
		}

		printf("OK\n");
	}

	BumpNdf(const Properties &props) : BSDF(props) {
		m_coeffFilename = props.getString("coeffFilename", "");
		m_boundFilename = props.getString("boundFilename", "");

		sigma_r1 = props.getFloat("roughness", 0.0001f);
		sigma_r2 = props.getFloat("roughness", 0.0001f);
		TILES = props.getFloat("TILES", 1.0f);
		NORMAL_MAP_RESOLUTION = props.getInteger("resolution", 1024);
		max_radius = props.getInteger("maxRadius", 1 << 30);
		MAX_ITER = props.getInteger("maxIter", 2);
	}

	BumpNdf(Stream *stream, InstanceManager *manager)
			: BSDF(stream, manager) {
		m_nested = static_cast<BSDF *>(manager->getInstance(stream));
		m_normal = static_cast<Texture *>(manager->getInstance(stream));

		sigma_r1 = stream->readFloat();
		sigma_r2= stream->readFloat();
		TILES = stream->readFloat();
		NORMAL_MAP_RESOLUTION = stream->readSize();
		max_radius = stream->readSize();
		MAX_ITER = stream->readSize();

		configure();
	}

	void configure() {
		if (!m_nested)
			Log(EError, "A child BSDF instance is required");
		if (!m_normal)
			Log(EError, "A displacement texture must be specified");

		m_components.clear();
		for (int i=0; i<m_nested->getComponentCount(); ++i)
			m_components.push_back(m_nested->getType(i) | ESpatiallyVarying | EAnisotropic);

		m_usesRayDifferentials = true;

		BSDF::configure();

		HIGHEST_LEVEL = (int) (log(double(NORMAL_MAP_RESOLUTION)) / log(2.0) + 0.5);

		read_coeff();
		read_bound();
		generateMipmap();
	}

	void serialize(Stream *stream, InstanceManager *manager) const {
		BSDF::serialize(stream, manager);

		manager->serialize(stream, m_nested.get());
		manager->serialize(stream, m_normal.get());
		
		stream->writeFloat(sigma_r1);
		stream->writeFloat(sigma_r2);
		stream->writeFloat(TILES);
		stream->writeSize(NORMAL_MAP_RESOLUTION);
		stream->writeSize(max_radius);
		stream->writeSize(MAX_ITER);
	}

	void addChild(const std::string &name, ConfigurableObject *child) {
		if (child->getClass()->derivesFrom(MTS_CLASS(BSDF))) {
			if (m_nested != NULL)
				Log(EError, "Only a single nested BSDF can be added!");
			m_nested = static_cast<BSDF *>(child);
		} else if (child->getClass()->derivesFrom(MTS_CLASS(Texture))) {
			if (m_normal != NULL)
				Log(EError, "Only a single displacement texture can be specified!");
			m_normal = static_cast<Texture *>(child);
		} else {
			BSDF::addChild(name, child);
		}
	}

	double coeff_poly(const double mu, const double sigma) const {
		return 1.0 / (sigma * sqrt(2.0 * M_PI));
	}

	double coeff_exp(const double mu, const double sigma) const {
		return -0.5 / pow((double)sigma, 2.0);
	}

	double erf(const double x) const {
	    if (x < 0) return erf(-x) * (-1);

	    double p = 0.3275911, a1 = 0.254829592, a2 = -0.284496736, a3 = 1.421413741, a4 = -1.453152027, a5 = 1.061405429;
	    double t = 1.0 / (1.0 + p * x);
	    return 1.0 - (a1 * t + a2 * t * t + a3 * pow(t, 3.0) + a4 * pow(t, 4.0) + a5 * pow(t, 5.0)) * exp(-x * x);
	}

	double max(const double a, const double b) const {
		return a > b ? a : b;
	}

	double min(const double a, const double b) const {
		return a > b ? b : a;
	}

	/* Calculate int(exp(a * x ^ 2 + b * x + c) * erf(d * x + e), x, x_start, x_end) */
	double exp_erf(double a, double b, double c, double d, double e, double x_start, double x_end) const {
	    if (d == 0.0) {
		    double A = sqrt(M_PI) / (2.0 * sqrt(-a));
			double B = exp(-b * b / (4.0 * a) + c);
			double t1 = (b / 2.0 + a * x_start) / sqrt(-a);
			double t2 = (b / 2.0 + a * x_end) / sqrt(-a);
			double C = erf(t1) - erf(t2);
			double int_exp = A * B * C;
			double ans = erf(e) * int_exp;
		    return ans;
	    }

	   	double p = d * x_start + e;
	    double q = d * x_end + e;

	    int flag = 1;
	    if (p > q) {flag = -1; swap(p, q);}
	    
	    double ret_val = 0;
	    for (int i = 0; i < ERF_SEG_NUM; i++) {
	        double int_start = max(ERF_SEG_COORD[i], p);
	        double int_end = min(ERF_SEG_COORD[i + 1], q);
	        if (int_start >= int_end) continue;
	        int_start = (int_start - e) / d;
	        int_end = (int_end - e) / d;
	        
	        double dd = ERF_SEG_COEFF[i][0] * d * d;
	        double ee = ERF_SEG_COEFF[i][0] * 2 * d * e + ERF_SEG_COEFF[i][1] * d;
	        double ff = ERF_SEG_COEFF[i][0] * e * e + ERF_SEG_COEFF[i][1] * e + ERF_SEG_COEFF[i][2];

			double f1 = sqrt(M_PI) * (4 * a * a * ff - 2 * a * (b * ee + dd) + b * b * dd) * (erf((2 * a * int_start + b) / (2 * sqrt(-a))) - erf((2 * a * int_end + b) / (2 * sqrt(-a))));
		    double f2 = 2 * sqrt(-a) * (exp(pow(2 * a * int_end + b, 2.0) / (4 * a)) * (2 * a * (dd * int_end + ee) - b * dd) - exp(pow(2 * a * int_start + b, 2.0) / (4 * a)) * (2 * a * (dd * int_start + ee) - b * dd));
		    ret_val += 1.0f / (8 * pow(-a, 2.5)) * exp(c - b * b / (4 * a)) * (f1 + f2);
	        // ret_val += (ee*exp(c+b*int_end+a*(int_end*int_end))*(1.0/2.0)-ee*exp(c+b*int_start+a*(int_start*int_start))*(1.0/2.0)+dd*int_end*exp(c+b*int_end+a*(int_end*int_end))*(1.0/2.0)-dd*int_start*exp(c+b*int_start+a*(int_start*int_start))*(1.0/2.0))/a-1.0/(a*a)*b*dd*(exp(c+b*int_end+a*(int_end*int_end))-exp(c+b*int_start+a*(int_start*int_start)))*(1.0/4.0)-sqrt(3.141592653589793)*1.0/pow(-a,3.0L/2.0L)*exp((a*c-(b*b)*(1.0/4.0))/a)*(erf(1.0/sqrt(-a)*(b+a*int_end*2.0)*(1.0/2.0))-erf(1.0/sqrt(-a)*(b+a*int_start*2.0)*(1.0/2.0)))*(dd+b*ee)*(1.0/4.0)-sqrt(3.141592653589793)*1.0/sqrt(-a)*ff*exp((a*c-(b*b)*(1.0/4.0))/a)*(erf(1.0/sqrt(-a)*(b*1.0+a*int_end*2.0*1.0)*(1.0/2.0))-erf(1.0/sqrt(-a)*(b*1.0+a*int_start*2.0*1.0)*(1.0/2.0)))*5.0E-1*1.0-sqrt(3.141592653589793)*1.0/pow(-a,5.0L/2.0L)*(b*b)*dd*exp((a*c-(b*b)*(1.0/4.0))/a)*(erf(1.0/sqrt(-a)*(b+a*int_end*2.0)*(1.0/2.0))-erf(1.0/sqrt(-a)*(b+a*int_start*2.0)*(1.0/2.0)))*(1.0/8.0);
	    }
	    ret_val *= flag;

	    return ret_val;
	}

	/* Calculate the integral over a sub-texel using two triangles */
	double int_quad(double sigma_p1, double sigma_p2, double x1a, double x1b, double x2a, double x2b, double x1, double x2, double y1, double y2) const {
		double coeff = coeff_poly(x1, sigma_p1) * coeff_poly(x2, sigma_p2) * coeff_poly(y1, sigma_r1) * coeff_poly(y2, sigma_r2);

	    double cx1 = coeff_exp(x1, sigma_p1);
	    double cx2 = coeff_exp(x2, sigma_p2);
	    double cy1 = coeff_exp(y1, sigma_r1);
	    double cy2 = coeff_exp(y2, sigma_r2);

	    double A[3][3], B[3];
	    memset(A, 0, sizeof(A));
	    memset(B, 0, sizeof(B));

	    double ret[2] = {0};

	    for (int i = 0; i < 2; i++)
		{
			if (i == 0) {
			    pair<double, double> N1 = getNormal(x1a, x2a);
			    pair<double, double> N2 = getNormal(x1b, x2a);
			    pair<double, double> N3 = getNormal(x1a, x2b);

			    double Y1[3] = {0.0, N1.first, N1.second};
			    double Y2[3] = {0.0, N2.first, N2.second};
			    double Y3[3] = {0.0, N3.first, N3.second};

			    A[1][1] = (Y1[1] - Y2[1]) / (x1a - x1b);
			    A[1][2] = (Y1[1] - Y3[1]) / (x2a - x2b);
			    A[2][1] = (Y1[2] - Y2[2]) / (x1a - x1b);
			    A[2][2] = (Y1[2] - Y3[2]) / (x2a - x2b);
			    B[1] = Y1[1] - (A[1][1] * x1a + A[1][2] * x2a);
			    B[2] = Y1[2] - (A[2][1] * x1a + A[2][2] * x2a);
			} else {
			    pair<double, double> N1 = getNormal(x1b, x2b);
			    pair<double, double> N2 = getNormal(x1b, x2a);
			    pair<double, double> N3 = getNormal(x1a, x2b);

			    double Y1[3] = {0.0, N1.first, N1.second};
			    double Y2[3] = {0.0, N2.first, N2.second};
			    double Y3[3] = {0.0, N3.first, N3.second};

			    A[1][1] = (Y3[1] - Y1[1]) / (x1a - x1b);
			    A[1][2] = (Y2[1] - Y1[1]) / (x2a - x2b);
			    A[2][1] = (Y3[2] - Y1[2]) / (x1a - x1b);
			    A[2][2] = (Y2[2] - Y1[2]) / (x2a - x2b);
			    B[1] = Y1[1] - (A[1][1] * x1b + A[1][2] * x2b);
			    B[2] = Y1[2] - (A[2][1] * x1b + A[2][2] * x2b);
			}

		    double a = cx1 + cy1 * A[1][1] * A[1][1] + cy2 * A[2][1] * A[2][1];
		    double b = cy1 * 2 * A[1][1] * A[1][2] + cy2 * 2 * A[2][1] * A[2][2];
		    double c = cx2 + cy1 * A[1][2] * A[1][2] + cy2 * A[2][2] * A[2][2];
		    double d = cx1 * 2 * (-x1) + cy1 * 2 * A[1][1] * (B[1] - y1) + cy2 * 2 * A[2][1] * (B[2] - y2);
		    double e = cx2 * 2 * (-x2) + cy1 * 2 * A[1][2] * (B[1] - y1) + cy2 * 2 * A[2][2] * (B[2] - y2);
		    double f = cx1 * x1 * x1 + cx2 * x2 * x2 + cy1 * (B[1] - y1) * (B[1] - y1) + cy2 * (B[2] - y2) * (B[2] - y2);

		    double ff = sqrt(M_PI) / (2 * sqrt(-c));
		    double aa = -(b * b - 4 * a * c) / (4 * c);
		    double bb = -(2 * b * e - 4 * c * d) / (4 * c);
		    double cc = -(e * e) / (4 * c) + f;

		    double dd1, ee1, dd2, ee2;
		    if (i == 0) {
			    dd2 = (b - (x2b - x2a) / (x1b - x1a) * 2 * c) / (2 * sqrt(-c));
			    ee2 = (e + 2 * c * x2a + x1b * (x2b - x2a) / (x1b - x1a) * 2 * c) / (2 * sqrt(-c));
				dd1 = b / (2 * sqrt(-c));
				ee1 = (e + 2 * c * x2a) / (2 * sqrt(-c));
		    } else {
				dd2 = b / (2 * sqrt(-c));
				ee2 = (e + 2 * c * x2b) / (2 * sqrt(-c));
			    dd1 = (b - (x2b - x2a) / (x1b - x1a) * 2 * c) / (2 * sqrt(-c));
			    ee1 = (e + 2 * c * x2a + x1b * (x2b - x2a) / (x1b - x1a) * 2 * c) / (2 * sqrt(-c));
		    }

		    ret[i] = ff * exp_erf(aa, bb, cc, dd1, ee1, x1a, x1b) - ff * exp_erf(aa, bb, cc, dd2, ee2, x1a, x1b);
		}
		return coeff * (ret[0] + ret[1]);
	}

	bool range_intersect(const double a1, const double a2, const double b1, const double b2) const {
		return (a1 >= b1 && a1 <= b2) ||
			   (a2 >= b1 && a2 <= b2) ||
			   (b1 >= a1 && b1 <= a2) ||
			   (b2 >= a1 && b2 <= a2);
	}
	
	/* Get sub-texel's (n1, n2) bound */
	void get_bound(double &max_n1, double &max_n2, double &min_n1, double &min_n2, double x1a, double x1b, double x2a, double x2b, int iters) const {
		max_n1 = max_n2 = -1e10;
		min_n1 = min_n2 = 1e10;

		/* We don't have acceleration structure for sub-texels, so we loop all of them */
		int max_segs = (1 << iters);
		for (int i = 0; i <= max_segs; i++) {
			for (int j = 0; j <= max_segs; j++) {
				float x1 = (float) (i) / max_segs * (x1b - x1a) + x1a;
				float x2 = (float) (j) / max_segs * (x2b - x2a) + x2a;
				
				pair<float, float> n = getNormal(x1, x2);
				
				max_n1 = max(max_n1, n.first);
				min_n1 = min(min_n1, n.first);
				max_n2 = max(max_n2, n.second);
				min_n2 = min(min_n2, n.second);
			}
		}
	}

	/* Subdivide texel into sub-texels */
	double int_quadx(int max_iter, int cur_iter, double sigma_p1, double sigma_p2, double x1a, double x1b, double x2a, double x2b, double x1, double x2, double y1, double y2) const {
		/* Stopping criteria: out of roughness range */
		if (cur_iter != 0) {
			double max_n1, max_n2, min_n1, min_n2;
			get_bound(max_n1, max_n2, min_n1, min_n2, x1a, x1b, x2a, x2b, max_iter - cur_iter);
			if (!range_intersect(min_n1, max_n1, y1 - sigma_r1 * 5.0, y1 + sigma_r1 * 5.0)) return 0.0;
			if (!range_intersect(min_n2, max_n2, y2 - sigma_r2 * 5.0, y2 + sigma_r2 * 5.0)) return 0.0;
		}

		/* Special case: don't want to subdivide */
		if (max_iter == 0) {
			return int_quad(sigma_p1, sigma_p2, x1a, x1b, x2a, x2b, x1, x2, y1, y2);
		}

		/* Stopping criteria: reached maximum iterations, perform integral per sub-texel */
		double ans1, ans2, ans3, ans4;
		if (cur_iter >= max_iter) {
			ans1 = int_quad(sigma_p1, sigma_p2, x1a, (x1a + x1b) / 2.0, x2a, (x2a + x2b) / 2.0, x1, x2, y1, y2);
			ans2 = int_quad(sigma_p1, sigma_p2, (x1a + x1b) / 2.0, x1b, x2a, (x2a + x2b) / 2.0, x1, x2, y1, y2);
			ans3 = int_quad(sigma_p1, sigma_p2, x1a, (x1a + x1b) / 2.0, (x2a + x2b) / 2.0, x2b, x1, x2, y1, y2);
			ans4 = int_quad(sigma_p1, sigma_p2, (x1a + x1b) / 2.0, x1b, (x2a + x2b) / 2.0, x2b, x1, x2, y1, y2);
			return ans1 + ans2 + ans3 + ans4;
		}

		/* Or else, we subdivide */
		ans1 = int_quadx(max_iter, cur_iter + 1, sigma_p1, sigma_p2, x1a, (x1a + x1b) / 2.0, x2a, (x2a + x2b) / 2.0, x1, x2, y1, y2);
		ans2 = int_quadx(max_iter, cur_iter + 1, sigma_p1, sigma_p2, (x1a + x1b) / 2.0, x1b, x2a, (x2a + x2b) / 2.0, x1, x2, y1, y2);
		ans3 = int_quadx(max_iter, cur_iter + 1, sigma_p1, sigma_p2, x1a, (x1a + x1b) / 2.0, (x2a + x2b) / 2.0, x2b, x1, x2, y1, y2);
		ans4 = int_quadx(max_iter, cur_iter + 1, sigma_p1, sigma_p2, (x1a + x1b) / 2.0, x1b, (x2a + x2b) / 2.0, x2b, x1, x2, y1, y2);
		return ans1 + ans2 + ans3 + ans4;
	}

	/* Search for texels that lie in both querying range and roughness range */
	double search_int_range(double sigma_p1, double sigma_p2, 
							int level, int p, int q, 
						 	int x1a, int x1b, int x2a, int x2b, 
							double x1, double x2, double y1, double y2,
							int max_iter) const {
		/* Stopping criteria 1: out of roughness range */
		if (!range_intersect(y1 - sigma_r1 * 5.0, y1 + sigma_r1 * 5.0, mipmap[level][p][q].y1a, mipmap[level][p][q].y1b)) return 0.0;
		if (!range_intersect(y2 - sigma_r2 * 5.0, y2 + sigma_r2 * 5.0, mipmap[level][p][q].y2a, mipmap[level][p][q].y2b)) return 0.0;

		/* Stopping criteria 2: reached a single texel */
		if (level == 0) {
			return int_quadx(max_iter, 0, sigma_p1, sigma_p2, (p + 0.0) / NORMAL_MAP_RESOLUTION, (p + 1.0) / NORMAL_MAP_RESOLUTION, (q + 0.0) / NORMAL_MAP_RESOLUTION, (q + 1.0) / NORMAL_MAP_RESOLUTION, x1, x2, y1, y2);
		}

		/* Stopping criteria 3: out of querying range */
		if (!range_intersect(p << level, ((p + 1) << level) - 1, x1a, x1b)) return 0.0;
		if (!range_intersect(q << level, ((q + 1) << level) - 1, x2a, x2b)) return 0.0;

		/* Or else, we subdivide searching range */		
		double aa = search_int_range(sigma_p1, sigma_p2, level - 1, p * 2 + 0, q * 2 + 0, x1a, x1b, x2a, x2b, x1, x2, y1, y2, max_iter);
		double bb = search_int_range(sigma_p1, sigma_p2, level - 1, p * 2 + 1, q * 2 + 0, x1a, x1b, x2a, x2b, x1, x2, y1, y2, max_iter);
		double cc = search_int_range(sigma_p1, sigma_p2, level - 1, p * 2 + 1, q * 2 + 1, x1a, x1b, x2a, x2b, x1, x2, y1, y2, max_iter);
		double dd = search_int_range(sigma_p1, sigma_p2, level - 1, p * 2 + 0, q * 2 + 1, x1a, x1b, x2a, x2b, x1, x2, y1, y2, max_iter);
		return aa + bb + cc + dd;
	}

	/* Calculate ndf given the half vector */
	double ndf(const Intersection &its, double y1, double y2) const {
		/* Calculate querying center and size */
		double du = fabs(its.dudx) + fabs(its.dudy);
		double dv = fabs(its.dvdx) + fabs(its.dvdy);

		double sigma_p1 = dv / 2.0;
		double sigma_p2 = du / 2.0;
		double x1 = its.uv.y; // In Mitsuba, u and v are along column and row, respectively. So we swap them.
		double x2 = its.uv.x;

		sigma_p1 *= TILES;
		sigma_p2 *= TILES;
		x1 *= TILES;
		x2 *= TILES;

		/* Constrain querying size in case extremely large size encountered */
		if (sigma_p1 * 3.0 * NORMAL_MAP_RESOLUTION > max_radius)
			sigma_p1 = max_radius / 3.0 / NORMAL_MAP_RESOLUTION;
		if (sigma_p2 * 3.0 * NORMAL_MAP_RESOLUTION > max_radius)
			sigma_p2 = max_radius / 3.0 / NORMAL_MAP_RESOLUTION;

		/* For large querying size or large roughness, we use fewer triangles per texel */
		int max_iter = MAX_ITER; // 2 for 32 triangles, 1 for 8, 0 for 2
		if (sigma_p1 * sigma_p2 * NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION > 225) max_iter--;
		if (sigma_p1 * sigma_p2 * NORMAL_MAP_RESOLUTION * NORMAL_MAP_RESOLUTION > 900) max_iter--;
		if (sigma_r1 >= 0.02) max_iter--;
		if (sigma_r1 >= 0.1) max_iter--;
		max_iter = max(0, max_iter);

		sigma_p1 = max(sigma_p1, 1.0 / NORMAL_MAP_RESOLUTION);
		sigma_p2 = max(sigma_p2, 1.0 / NORMAL_MAP_RESOLUTION);

		/* Chop and shift the querying area manually to fit in range(0, N_M_R) */
		/* Since the acceleration structure don't repeat */
		double ret_val = 0.0;

		for (int i = floor(x1 - sigma_p1 * 3.0); i <= ceil(x1 + sigma_p1 * 3.0); i++) {
			for (int j = floor(x2 - sigma_p2 * 3.0); j <= ceil(x2 + sigma_p2 * 3.0); j++) {
				int x1a = floor(max((x1 - sigma_p1 * 3.0 - i) * NORMAL_MAP_RESOLUTION, 0));
				int x1b = ceil(min((x1 + sigma_p1 * 3.0 - i) * NORMAL_MAP_RESOLUTION, NORMAL_MAP_RESOLUTION - 1));
				if (x1a > x1b) continue;
				int x2a = floor(max((x2 - sigma_p2 * 3.0 - j) * NORMAL_MAP_RESOLUTION, 0));
				int x2b = ceil(min((x2 + sigma_p2 * 3.0 - j) * NORMAL_MAP_RESOLUTION, NORMAL_MAP_RESOLUTION - 1));
				if (x2a > x2b) continue;

				double x11 = x1 - i;
				double x22 = x2 - j;

				/* Perform quad tree search and compute for texels that lie in roughness range */
				ret_val += search_int_range(sigma_p1, sigma_p2, HIGHEST_LEVEL, 0, 0, x1a, x1b, x2a, x2b, x11, x22, y1, y2, max_iter);	
			}
		}

		return ret_val;
	}
	
	Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const {
		/* Calculate the half-vector */
		Vector H = normalize(bRec.wo + bRec.wi);

		/* Calculate the (n1, n2) for querying in local coordinate*/
		Float n1 = dot(H, bRec.its.toLocal(normalize(bRec.its.shFrame.s)));
		Float n2 = dot(H, bRec.its.toLocal(normalize(bRec.its.shFrame.t)));

		/* Calculate the ndf at (n1, n2) */
		Float ndf_term = ndf(bRec.its, n1, n2);

		/* Calculate the BSDF using our ndf and the rest of microfacet model */
		BSDFSamplingRecord gRec = bRec;
		gRec.sampledComponent = 0; // Used to store specify glossy (0) or diffuse (1)

		// return Spectrum(ndf_term);

		return m_nested->eval(gRec, measure) * max(ndf_term, 0.0);
	}

	Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const {
		return 0.0f;
	}

	Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const {
		return Spectrum(0.0f);
	}

	Spectrum sample(BSDFSamplingRecord &bRec, Float &pdf, const Point2 &sample) const {
		return Spectrum(0.0f);
	}

	Float getRoughness(const Intersection &its, int component) const {
		return m_nested->getRoughness(its, component);
	}

	std::string toString() const {
		std::ostringstream oss;
		oss << "BumpMap[" << endl
			<< "  id = \"" << getID() << "\"," << endl
			<< "  displacement = " << indent(m_normal->toString()) << endl
			<< "  nested = " << indent(m_nested->toString()) << endl
			<< "]";
		return oss.str();
	}

	Shader *createShader(Renderer *renderer) const;

	MTS_DECLARE_CLASS()
protected:
	ref<Texture> m_normal;
	ref<BSDF> m_nested;
	string m_coeffFilename, m_boundFilename;
	Float sigma_r1;
	Float sigma_r2;
	Float TILES;
	int max_radius;
	int MAX_ITER;
};

// ================ Hardware shader implementation ================

/**
 * This is a quite approximate version of the bump map model -- it likely
 * won't match the reference exactly, but it should be good enough for
 * preview purposes
 */
class BumpShaderNdf : public Shader {
public:
	BumpShaderNdf(Renderer *renderer, const BSDF *nested, const Texture *displacement)
		: Shader(renderer, EBSDFShader), m_nested(nested), m_normal(displacement) {
		m_nestedShader = renderer->registerShaderForResource(m_nested.get());
		m_normalShader = renderer->registerShaderForResource(m_normal.get());
	}

	bool isComplete() const {
		return m_nestedShader.get() != NULL;
	}

	void cleanup(Renderer *renderer) {
		renderer->unregisterShaderForResource(m_nested.get());
		renderer->unregisterShaderForResource(m_normal.get());
	}

	void putDependencies(std::vector<Shader *> &deps) {
		deps.push_back(m_nestedShader.get());
		deps.push_back(m_normalShader.get());
	}

	void generateCode(std::ostringstream &oss,
			const std::string &evalName,
			const std::vector<std::string> &depNames) const {
		oss << "vec3 " << evalName << "(vec2 uv, vec3 wi, vec3 wo) {" << endl
			<< "    float du = abs(dFdx(uv.x)), dv = abs(dFdx(uv.y));" << endl
			<< "    if (du == 0.0) du = 0.001;" << endl
			<< "    if (dv == 0.0) dv = 0.001;" << endl
			<< "    float displacement = " << depNames[1] << "(uv)[0];" << endl
			<< "    float displacementU = " << depNames[1] << "(uv + vec2(du, 0.0))[0];" << endl
			<< "    float displacementV = " << depNames[1] << "(uv + vec2(0.0, dv))[0];" << endl
			<< "    float dfdu = 0.0;//(displacementU - displacement)/du;" << endl
			<< "    float dfdv = 0.0;//(displacementV - displacement)/dv;" << endl
			<< "    vec3 dpdu = normalize(vec3(1.0, 0.0, dfdu));" << endl
			<< "    vec3 dpdv = vec3(0.0, 1.0, dfdv);" << endl
			<< "    dpdv = normalize(dpdv - dot(dpdu, dpdv)*dpdu);" << endl
			<< "    vec3 n = cross(dpdu, dpdv);" << endl
			<< "    wi = vec3(dot(wi, dpdu), dot(wi, dpdv), dot(wi, n));" << endl
			<< "    wo = vec3(dot(wo, dpdu), dot(wo, dpdv), dot(wo, n));" << endl
			<< "    return " << depNames[0] << "(uv, wi, wo);" << endl
			<< "}" << endl
			<< endl
			<< "vec3 " << evalName << "_diffuse(vec2 uv, vec3 wi, vec3 wo) {" << endl
			<< "    float du = abs(dFdx(uv.x)), dv = abs(dFdx(uv.y));" << endl
			<< "    if (du == 0.0) du = 0.001;" << endl
			<< "    if (dv == 0.0) dv = 0.001;" << endl
			<< "    float displacement = " << depNames[1] << "(uv)[0];" << endl
			<< "    float displacementU = " << depNames[1] << "(uv + vec2(du, 0.0))[0];" << endl
			<< "    float displacementV = " << depNames[1] << "(uv + vec2(0.0, dv))[0];" << endl
			<< "    float dfdu = 0.0;//(displacementU - displacement)/du;" << endl
			<< "    float dfdv = 0.0;//(displacementV - displacement)/dv;" << endl
			<< "    vec3 dpdu = normalize(vec3(1.0, 0.0, dfdu));" << endl
			<< "    vec3 dpdv = vec3(0.0, 1.0, dfdv);" << endl
			<< "    dpdv = normalize(dpdv - dot(dpdu, dpdv)*dpdu);" << endl
			<< "    vec3 n = cross(dpdu, dpdv);" << endl
			<< "    wi = vec3(dot(wi, dpdu), dot(wi, dpdv), dot(wi, n));" << endl
			<< "    wo = vec3(dot(wo, dpdu), dot(wo, dpdv), dot(wo, n));" << endl
			<< "    return " << depNames[0] << "_diffuse(uv, wi, wo);" << endl
			<< "}" << endl;
	}

	MTS_DECLARE_CLASS()
private:
	ref<const BSDF> m_nested;
	ref<const Texture> m_normal;
	ref<Shader> m_nestedShader;
	ref<Shader> m_normalShader;
};

Shader *BumpNdf::createShader(Renderer *renderer) const {
	return new BumpShaderNdf(renderer, m_nested.get(), m_normal.get());
}

MTS_IMPLEMENT_CLASS(BumpShaderNdf, false, Shader)
MTS_IMPLEMENT_CLASS_S(BumpNdf, false, BSDF)
MTS_EXPORT_PLUGIN(BumpNdf, "OH YEAH!!!");
MTS_NAMESPACE_END
